from typing import Dict, List, Optional

from phi.k8s.create.base import CreateK8sResource
from phi.k8s.enums.api_version import ApiVersion
from phi.k8s.enums.kind import Kind
from phi.k8s.resource.rbac_authorization_k8s_io.v1.cluster_role import (
    ClusterRole,
    PolicyRule,
)
from phi.k8s.create.common.labels import create_component_labels_dict
from phi.k8s.resource.meta.v1.object_meta import ObjectMeta


class CreateClusterRole(CreateK8sResource):
    cr_name: str
    app_name: str
    rules: Optional[List[PolicyRule]] = None
    namespace: Optional[str] = None
    labels: Optional[Dict[str, str]] = None

    def _create(self) -> ClusterRole:
        """Creates the ClusterRole resource"""

        cr_name = self.cr_name
        # logger.debug(f"Init ClusterRole resource: {cr_name}")

        cr_labels = create_component_labels_dict(
            component_name=cr_name,
            app_name=self.app_name,
            labels=self.labels,
        )

        cr_rules: List[PolicyRule] = (
            self.rules if self.rules else [PolicyRule(api_groups=["*"], resources=["*"], verbs=["*"])]
        )

        cr = ClusterRole(
            name=cr_name,
            api_version=ApiVersion.RBAC_AUTH_V1,
            kind=Kind.CLUSTERROLE,
            metadata=ObjectMeta(
                name=cr_name,
                namespace=self.namespace,
                labels=cr_labels,
            ),
            rules=cr_rules,
        )
        return cr



from typing import Optional, Union, List

from phi.k8s.app.superset.base import SupersetBase, RestartPolicy


class SupersetInit(SupersetBase):
    # -*- App Name
    name: str = "superset-init"

    # Command for the container
    entrypoint: Optional[Union[str, List]] = "/scripts/init-superset.sh"

    # -*- Deployment Configuration
    restart_policy: Optional[RestartPolicy] = RestartPolicy.ON_FAILURE



from os import getenv
from typing import Optional, Union, Dict, List

from httpx import Response

from phi.api.api import api, invalid_response
from phi.api.routes import ApiRoutes
from phi.api.schemas.conversation import (
    ConversationEventCreate,
    ConversationWorkspace,
    ConversationEventCreateResopnse,
    ConversationUpdate,
    ConversationSchema,
)
from phi.constants import WORKSPACE_ID_ENV_VAR, WORKSPACE_HASH_ENV_VAR
from phi.cli.settings import phi_cli_settings
from phi.utils.common import str_to_int
from phi.utils.log import logger


def upsert_conversation(conversation: ConversationUpdate) -> Optional[ConversationSchema]:
    if not phi_cli_settings.api_enabled:
        return None

    logger.debug("--o-o-- Conversation upsert")
    with api.Client() as api_client:
        try:
            workspace_id = str_to_int(getenv(WORKSPACE_ID_ENV_VAR))
            if workspace_id is None:
                return None
            workspace_hash = getenv(WORKSPACE_HASH_ENV_VAR)
            if workspace_hash is None:
                return None
            workspace = ConversationWorkspace(id_workspace=workspace_id, ws_hash=workspace_hash)

            r: Response = api_client.post(
                ApiRoutes.CONVERSATION_UPSERT,
                json={
                    "conversation": conversation.model_dump(exclude_none=True),
                    "workspace": workspace.model_dump(exclude_none=True),
                },
            )
            if invalid_response(r):
                return None

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return None

            conversation_schema: Optional[ConversationSchema] = ConversationSchema.model_validate(response_json)
            # logger.debug(f"Conversation: {conversation_schema}")
            return conversation_schema
        except Exception as e:
            logger.debug(f"Could not update conversation: {e}")
    return None


def log_conversation_event(conversation: ConversationEventCreate) -> Optional[ConversationEventCreateResopnse]:
    if not phi_cli_settings.api_enabled:
        return None

    logger.debug("--o-o-- Log conversation event")
    with api.Client() as api_client:
        try:
            workspace_id = str_to_int(getenv(WORKSPACE_ID_ENV_VAR))
            if workspace_id is None:
                return None
            workspace_hash = getenv(WORKSPACE_HASH_ENV_VAR)
            if workspace_hash is None:
                return None
            workspace = ConversationWorkspace(id_workspace=workspace_id, ws_hash=workspace_hash)

            r: Response = api_client.post(
                ApiRoutes.CONVERSATION_EVENT_CREATE,
                json={
                    "event": conversation.model_dump(exclude_none=True),
                    "workspace": workspace.model_dump(exclude_none=True),
                },
            )
            if invalid_response(r):
                return None

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return None

            conversation_event_response: ConversationEventCreateResopnse = (
                ConversationEventCreateResopnse.model_validate(response_json)
            )
            # logger.debug(f"Conversation response: {conversation_event_response}")
            return conversation_event_response
        except Exception as e:
            logger.debug(f"Could not log conversation event: {e}")
    return None



from typing import Dict, Any

from pydantic import BaseModel


class Ivfflat(BaseModel):
    lists: int = 100
    probes: int = 10
    dynamic_lists: bool = True
    configuration: Dict[str, Any] = {
        "maintenance_work_mem": "2GB",
    }


class HNSW(BaseModel):
    m: int = 16
    ef: int = 50
    ef_construction: int = 200
    configuration: Dict[str, Any] = {
        "maintenance_work_mem": "2GB",
    }



from typing import Dict, List, Optional
from typing_extensions import Literal

from phi.k8s.create.base import CreateK8sResource
from phi.k8s.create.apps.v1.deployment import CreateDeployment
from phi.k8s.create.common.port import CreatePort
from phi.k8s.enums.api_version import ApiVersion
from phi.k8s.enums.kind import Kind
from phi.k8s.enums.service_type import ServiceType
from phi.k8s.resource.core.v1.service import Service, ServicePort, ServiceSpec
from phi.k8s.create.common.labels import create_component_labels_dict
from phi.k8s.resource.meta.v1.object_meta import ObjectMeta


class CreateService(CreateK8sResource):
    service_name: str
    app_name: str
    namespace: Optional[str] = None
    service_account_name: Optional[str] = None
    service_type: Optional[ServiceType] = None
    # Deployment to expose using this service
    deployment: CreateDeployment
    # Ports to expose using this service
    ports: Optional[List[CreatePort]] = None
    labels: Optional[Dict[str, str]] = None
    annotations: Optional[Dict[str, str]] = None
    # If ServiceType == ClusterIP
    cluster_ip: Optional[str] = None
    cluster_ips: Optional[List[str]] = None
    # If ServiceType == ExternalName
    external_ips: Optional[List[str]] = None
    external_name: Optional[str] = None
    external_traffic_policy: Optional[Literal["Cluster", "Local"]] = None
    # If ServiceType == ServiceType.LoadBalancer
    health_check_node_port: Optional[int] = None
    internal_traffic_policy: Optional[str] = None
    load_balancer_class: Optional[str] = None
    load_balancer_ip: Optional[str] = None
    load_balancer_source_ranges: Optional[List[str]] = None
    allocate_load_balancer_node_ports: Optional[bool] = None
    # Only used to print the LoadBalancer DNS
    protocol: Optional[str] = None

    def _create(self) -> Service:
        """Creates a Service resource"""
        service_name = self.service_name
        # logger.debug(f"Init Service resource: {service_name}")

        service_labels = create_component_labels_dict(
            component_name=service_name,
            app_name=self.app_name,
            labels=self.labels,
        )

        target_pod_name = self.deployment.pod_name
        target_pod_labels = create_component_labels_dict(
            component_name=target_pod_name,
            app_name=self.app_name,
            labels=self.labels,
        )

        service_ports: List[ServicePort] = []
        if self.ports:
            for _port in self.ports:
                # logger.debug(f"Creating ServicePort for {_port}")
                if _port.service_port is not None:
                    service_ports.append(
                        ServicePort(
                            name=_port.name,
                            port=_port.service_port,
                            node_port=_port.node_port,
                            protocol=_port.protocol,
                            target_port=_port.target_port,
                        )
                    )

        service = Service(
            name=service_name,
            api_version=ApiVersion.CORE_V1,
            kind=Kind.SERVICE,
            metadata=ObjectMeta(
                name=service_name,
                namespace=self.namespace,
                labels=service_labels,
                annotations=self.annotations,
            ),
            spec=ServiceSpec(
                type=self.service_type,
                cluster_ip=self.cluster_ip,
                cluster_ips=self.cluster_ips,
                external_ips=self.external_ips,
                external_name=self.external_name,
                external_traffic_policy=self.external_traffic_policy,
                health_check_node_port=self.health_check_node_port,
                internal_traffic_policy=self.internal_traffic_policy,
                load_balancer_class=self.load_balancer_class,
                load_balancer_ip=self.load_balancer_ip,
                load_balancer_source_ranges=self.load_balancer_source_ranges,
                allocate_load_balancer_node_ports=self.allocate_load_balancer_node_ports,
                ports=service_ports,
                selector=target_pod_labels,
            ),
            protocol=self.protocol,
        )

        # logger.debug(
        #     f"Service {service_name}:\n{service.json(exclude_defaults=True, indent=2)}"
        # )
        return service



from typing import Any, Dict, Optional

from phi.k8s.create.base import CreateK8sResource
from phi.k8s.enums.api_version import ApiVersion
from phi.k8s.enums.kind import Kind
from phi.k8s.create.apiextensions_k8s_io.v1.custom_resource_definition import (
    CreateCustomResourceDefinition,
)
from phi.k8s.resource.apiextensions_k8s_io.v1.custom_object import (
    CustomObject,
)
from phi.k8s.create.common.labels import create_component_labels_dict
from phi.k8s.resource.meta.v1.object_meta import ObjectMeta


class CreateCustomObject(CreateK8sResource):
    name: str
    app_name: str
    crd: CreateCustomResourceDefinition
    version: Optional[str] = None
    spec: Optional[Dict[str, Any]] = None
    namespace: Optional[str] = None
    service_account_name: Optional[str] = None
    labels: Optional[Dict[str, str]] = None

    def _create(self) -> CustomObject:
        """Creates a CustomObject resource."""
        # logger.debug(f"Creating CustomObject Resource: {group_name}")

        custom_object_name = self.name
        custom_object_labels = create_component_labels_dict(
            component_name=custom_object_name,
            app_name=self.app_name,
            labels=self.labels,
        )

        api_group_str: str = self.crd.group
        api_version_str: Optional[str] = None
        if self.version is not None and isinstance(self.version, str):
            api_version_str = self.version
        elif len(self.crd.versions) >= 1:
            api_version_str = self.crd.versions[0].name
        # api_version is required
        if api_version_str is None:
            raise ValueError(f"CustomObject ApiVersion invalid: {api_version_str}")

        plural: Optional[str] = self.crd.names.plural
        # plural is required
        if plural is None:
            raise ValueError(f"CustomResourceDefinition plural invalid: {plural}")

        # validate api_group_str and api_version_str
        api_group_version_str = "{}/{}".format(api_group_str, api_version_str)
        api_version_enum = None
        try:
            api_version_enum = ApiVersion.from_str(api_group_version_str)
        except NotImplementedError:
            raise NotImplementedError(f"{api_group_version_str} is not a supported API version")

        kind_str: str = self.crd.names.kind
        kind_enum = None
        try:
            kind_enum = Kind.from_str(kind_str)
        except NotImplementedError:
            raise NotImplementedError(f"{kind_str} is not a supported Kind")

        custom_object = CustomObject(
            name=custom_object_name,
            api_version=api_version_enum,
            kind=kind_enum,
            metadata=ObjectMeta(
                name=custom_object_name,
                namespace=self.namespace,
                labels=custom_object_labels,
            ),
            group=api_group_str,
            version=api_version_str,
            plural=plural,
            spec=self.spec,
        )

        # logger.debug(
        #     f"CustomObject {custom_object_name}:\n{custom_object.json(exclude_defaults=True, indent=2)}"
        # )
        return custom_object



from phi.aws.resource.eks.addon import EksAddon
from phi.aws.resource.eks.cluster import EksCluster
from phi.aws.resource.eks.fargate_profile import EksFargateProfile
from phi.aws.resource.eks.node_group import EksNodeGroup
from phi.aws.resource.eks.kubeconfig import EksKubeconfig



from typing import Optional, Dict, Any

from phi.app.db_app import DbApp
from phi.k8s.app.base import (
    K8sApp,
    AppVolumeType,
    ContainerContext,
    ServiceType,  # noqa: F401
    RestartPolicy,  # noqa: F401
    ImagePullPolicy,  # noqa: F401
)


class Redis(K8sApp, DbApp):
    # -*- App Name
    name: str = "redis"

    # -*- Image Configuration
    image_name: str = "redis"
    image_tag: str = "7.2.0"

    # -*- App Ports
    # Open a container port if open_port=True
    open_port: bool = True
    port_number: int = 6379
    # Port name for the opened port
    container_port_name: str = "redis"

    # -*- Service Configuration
    create_service: bool = True

    # -*- Redis Volume
    # Create a volume for redis storage
    create_volume: bool = True
    volume_type: AppVolumeType = AppVolumeType.EmptyDir
    # Path to mount the volume inside the container
    # should be the parent directory of pgdata defined above
    volume_container_path: str = "/data"
    # -*- If volume_type is AwsEbs
    ebs_volume: Optional[Any] = None
    # Add NodeSelectors to Pods, so they are scheduled in the same region and zone as the ebs_volume
    schedule_pods_in_ebs_topology: bool = True

    # -*- Redis Configuration
    # Provide REDIS_PASSWORD as redis_password or REDIS_PASSWORD in secrets_file
    redis_password: Optional[str] = None
    # Provide REDIS_SCHEMA as redis_schema or REDIS_SCHEMA in secrets_file
    redis_schema: Optional[str] = None
    redis_driver: str = "redis"
    logging_level: str = "debug"

    def get_db_password(self) -> Optional[str]:
        return self.redis_password or self.get_secret_from_file("REDIS_PASSWORD")

    def get_db_database(self) -> Optional[str]:
        return self.redis_schema or self.get_secret_from_file("REDIS_SCHEMA")

    def get_db_driver(self) -> Optional[str]:
        return self.redis_driver

    def get_db_host(self) -> Optional[str]:
        return self.get_service_name()

    def get_db_port(self) -> Optional[int]:
        return self.get_service_port()

    def get_db_connection(self) -> Optional[str]:
        password = self.get_db_password()
        password_str = f"{password}@" if password else ""
        schema = self.get_db_database()
        driver = self.get_db_driver()
        host = self.get_db_host()
        port = self.get_db_port()
        return f"{driver}://{password_str}{host}:{port}/{schema}"

    def get_db_connection_local(self) -> Optional[str]:
        password = self.get_db_password()
        password_str = f"{password}@" if password else ""
        schema = self.get_db_database()
        driver = self.get_db_driver()
        host = self.get_db_host_local()
        port = self.get_db_port_local()
        return f"{driver}://{password_str}{host}:{port}/{schema}"

    def get_container_env(self, container_context: ContainerContext) -> Dict[str, str]:
        # Container Environment
        container_env: Dict[str, str] = self.container_env or {}

        # Update the container env using env_file
        env_data_from_file = self.get_env_file_data()
        if env_data_from_file is not None:
            container_env.update({k: str(v) for k, v in env_data_from_file.items() if v is not None})

        # Update the container env with user provided env_vars
        # this overwrites any existing variables with the same key
        if self.env_vars is not None and isinstance(self.env_vars, dict):
            container_env.update({k: str(v) for k, v in self.env_vars.items() if v is not None})

        return container_env



try:
    from sqlalchemy.orm import DeclarativeBase
except ImportError:
    raise ImportError("`sqlalchemy` not installed")


class BaseTable(DeclarativeBase):
    """
    Base class for SQLAlchemy model definitions.

    https://docs.sqlalchemy.org/en/20/orm/mapping_api.html#sqlalchemy.orm.DeclarativeBase
    https://fastapi.tiangolo.com/tutorial/sql-databases/#create-a-base-class
    """

    pass



from phi.docker.app.airflow.base import AirflowBase, AirflowLogsVolumeType, ContainerContext
from phi.docker.app.airflow.webserver import AirflowWebserver
from phi.docker.app.airflow.scheduler import AirflowScheduler
from phi.docker.app.airflow.worker import AirflowWorker
from phi.docker.app.airflow.flower import AirflowFlower



from kubernetes.client.models.v1_volume_node_affinity import V1VolumeNodeAffinity

from phi.k8s.resource.base import K8sObject
from phi.k8s.resource.core.v1.node_selector import NodeSelector


class VolumeNodeAffinity(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#volumenodeaffinity-v1-core
    """

    resource_type: str = "VolumeNodeAffinity"

    # Required specifies hard node constraints that must be met.
    required: NodeSelector

    def get_k8s_object(
        self,
    ) -> V1VolumeNodeAffinity:
        # Return a V1VolumeNodeAffinity object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_volume_node_affinity.py
        _v1_volume_node_affinity = V1VolumeNodeAffinity(required=self.required.get_k8s_object())
        return _v1_volume_node_affinity



from typing import Optional

from pydantic import BaseModel

from phi.api.schemas.workspace import WorkspaceSchema


class ContainerContext(BaseModel):
    workspace_name: str
    # Path to the workspace directory inside the container
    workspace_root: str
    # Path to the workspace parent directory inside the container
    workspace_parent: str
    scripts_dir: Optional[str] = None
    storage_dir: Optional[str] = None
    workflows_dir: Optional[str] = None
    workspace_dir: Optional[str] = None
    workspace_schema: Optional[WorkspaceSchema] = None
    requirements_file: Optional[str] = None



from phi.k8s.app.superset.base import (
    SupersetBase,
    AppVolumeType,
    ContainerContext,
    ServiceType,
    RestartPolicy,
    ImagePullPolicy,
)
from phi.k8s.app.superset.webserver import SupersetWebserver
from phi.k8s.app.superset.init import SupersetInit
from phi.k8s.app.superset.worker import SupersetWorker
from phi.k8s.app.superset.worker_beat import SupersetWorkerBeat



from typing import List, Optional, Union, Tuple

from phi.app.group import AppGroup
from phi.resource.group import ResourceGroup
from phi.docker.app.base import DockerApp
from phi.docker.app.context import DockerBuildContext
from phi.docker.api_client import DockerApiClient
from phi.docker.resource.base import DockerResource
from phi.infra.resources import InfraResources
from phi.workspace.settings import WorkspaceSettings
from phi.utils.log import logger


class DockerResources(InfraResources):
    env: str = "dev"
    network: str = "phi"
    # URL for the Docker server. For example, unix:///var/run/docker.sock or tcp://127.0.0.1:1234
    base_url: Optional[str] = None

    apps: Optional[List[Union[DockerApp, AppGroup]]] = None
    resources: Optional[List[Union[DockerResource, ResourceGroup]]] = None

    # -*- Cached Data
    _api_client: Optional[DockerApiClient] = None

    @property
    def docker_client(self) -> DockerApiClient:
        if self._api_client is None:
            self._api_client = DockerApiClient(base_url=self.base_url)
        return self._api_client

    def create_resources(
        self,
        group_filter: Optional[str] = None,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        dry_run: Optional[bool] = False,
        auto_confirm: Optional[bool] = False,
        force: Optional[bool] = None,
    ) -> Tuple[int, int]:
        from phi.cli.console import print_info, print_heading, confirm_yes_no
        from phi.docker.resource.types import DockerContainer, DockerResourceInstallOrder

        logger.debug("-*- Creating DockerResources")
        # Build a list of DockerResources to create
        resources_to_create: List[DockerResource] = []
        if self.resources is not None:
            for r in self.resources:
                if isinstance(r, ResourceGroup):
                    resources_from_resource_group = r.get_resources()
                    if len(resources_from_resource_group) > 0:
                        for resource_from_resource_group in resources_from_resource_group:
                            if isinstance(resource_from_resource_group, DockerResource):
                                resource_from_resource_group.set_workspace_settings(
                                    workspace_settings=self.workspace_settings
                                )
                                if resource_from_resource_group.group is None and self.name is not None:
                                    resource_from_resource_group.group = self.name
                                if resource_from_resource_group.should_create(
                                    group_filter=group_filter,
                                    name_filter=name_filter,
                                    type_filter=type_filter,
                                ):
                                    resources_to_create.append(resource_from_resource_group)
                elif isinstance(r, DockerResource):
                    r.set_workspace_settings(workspace_settings=self.workspace_settings)
                    if r.group is None and self.name is not None:
                        r.group = self.name
                    if r.should_create(
                        group_filter=group_filter,
                        name_filter=name_filter,
                        type_filter=type_filter,
                    ):
                        r.set_workspace_settings(workspace_settings=self.workspace_settings)
                        resources_to_create.append(r)

        # Build a list of DockerApps to create
        apps_to_create: List[DockerApp] = []
        if self.apps is not None:
            for app in self.apps:
                if isinstance(app, AppGroup):
                    apps_from_app_group = app.get_apps()
                    if len(apps_from_app_group) > 0:
                        for app_from_app_group in apps_from_app_group:
                            if isinstance(app_from_app_group, DockerApp):
                                if app_from_app_group.group is None and self.name is not None:
                                    app_from_app_group.group = self.name
                                if app_from_app_group.should_create(group_filter=group_filter):
                                    apps_to_create.append(app_from_app_group)
                elif isinstance(app, DockerApp):
                    if app.group is None and self.name is not None:
                        app.group = self.name
                    if app.should_create(group_filter=group_filter):
                        apps_to_create.append(app)

        # Get the list of DockerResources from the DockerApps
        if len(apps_to_create) > 0:
            logger.debug(f"Found {len(apps_to_create)} apps to create")
            for app in apps_to_create:
                app.set_workspace_settings(workspace_settings=self.workspace_settings)
                app_resources = app.get_resources(build_context=DockerBuildContext(network=self.network))
                if len(app_resources) > 0:
                    # If the app has dependencies, add the resources from the
                    # dependencies first to the list of resources to create
                    if app.depends_on is not None:
                        for dep in app.depends_on:
                            if isinstance(dep, DockerApp):
                                dep.set_workspace_settings(workspace_settings=self.workspace_settings)
                                dep_resources = dep.get_resources(
                                    build_context=DockerBuildContext(network=self.network)
                                )
                                if len(dep_resources) > 0:
                                    for dep_resource in dep_resources:
                                        if isinstance(dep_resource, DockerResource):
                                            resources_to_create.append(dep_resource)
                    # Add the resources from the app to the list of resources to create
                    for app_resource in app_resources:
                        if isinstance(app_resource, DockerResource) and app_resource.should_create(
                            group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                        ):
                            resources_to_create.append(app_resource)

        # Sort the DockerResources in install order
        resources_to_create.sort(key=lambda x: DockerResourceInstallOrder.get(x.__class__.__name__, 5000))

        # Deduplicate DockerResources
        deduped_resources_to_create: List[DockerResource] = []
        for r in resources_to_create:
            if r not in deduped_resources_to_create:
                deduped_resources_to_create.append(r)

        # Implement dependency sorting
        final_docker_resources: List[DockerResource] = []
        logger.debug("-*- Building DockerResources dependency graph")
        for docker_resource in deduped_resources_to_create:
            # Logic to follow if resource has dependencies
            if docker_resource.depends_on is not None:
                # Add the dependencies before the resource itself
                for dep in docker_resource.depends_on:
                    if isinstance(dep, DockerResource):
                        if dep not in final_docker_resources:
                            logger.debug(f"-*- Adding {dep.name}, dependency of {docker_resource.name}")
                            final_docker_resources.append(dep)

                # Add the resource to be created after its dependencies
                if docker_resource not in final_docker_resources:
                    logger.debug(f"-*- Adding {docker_resource.name}")
                    final_docker_resources.append(docker_resource)
            else:
                # Add the resource to be created if it has no dependencies
                if docker_resource not in final_docker_resources:
                    logger.debug(f"-*- Adding {docker_resource.name}")
                    final_docker_resources.append(docker_resource)

        # Track the total number of DockerResources to create for validation
        num_resources_to_create: int = len(final_docker_resources)
        num_resources_created: int = 0
        if num_resources_to_create == 0:
            return 0, 0

        if dry_run:
            print_heading("--**- Docker resources to create:")
            for resource in final_docker_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info(f"\nNetwork: {self.network}")
            print_info(f"Total {num_resources_to_create} resources")
            return 0, 0

        # Validate resources to be created
        if not auto_confirm:
            print_heading("\n--**-- Confirm resources to create:")
            for resource in final_docker_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info(f"\nNetwork: {self.network}")
            print_info(f"Total {num_resources_to_create} resources")
            confirm = confirm_yes_no("\nConfirm deploy")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping create")
                print_info("-*-")
                return 0, 0

        for resource in final_docker_resources:
            print_info(f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}")
            if force is True:
                resource.force = True
            if isinstance(resource, DockerContainer):
                if resource.network is None and self.network is not None:
                    resource.network = self.network
            # logger.debug(resource)
            try:
                _resource_created = resource.create(docker_client=self.docker_client)
                if _resource_created:
                    num_resources_created += 1
                else:
                    if self.workspace_settings is not None and not self.workspace_settings.continue_on_create_failure:
                        return num_resources_created, num_resources_to_create
            except Exception as e:
                logger.error(f"Failed to create {resource.get_resource_type()}: {resource.get_resource_name()}")
                logger.error(e)
                logger.error("Please fix and try again...")

        print_heading(f"\n--**-- Resources created: {num_resources_created}/{num_resources_to_create}")
        if num_resources_to_create != num_resources_created:
            logger.error(
                f"Resources created: {num_resources_created} do not match resources required: {num_resources_to_create}"
            )  # noqa: E501
        return num_resources_created, num_resources_to_create

    def delete_resources(
        self,
        group_filter: Optional[str] = None,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        dry_run: Optional[bool] = False,
        auto_confirm: Optional[bool] = False,
        force: Optional[bool] = None,
    ) -> Tuple[int, int]:
        from phi.cli.console import print_info, print_heading, confirm_yes_no
        from phi.docker.resource.types import DockerContainer, DockerResourceInstallOrder

        logger.debug("-*- Deleting DockerResources")
        # Build a list of DockerResources to delete
        resources_to_delete: List[DockerResource] = []
        if self.resources is not None:
            for r in self.resources:
                if isinstance(r, ResourceGroup):
                    resources_from_resource_group = r.get_resources()
                    if len(resources_from_resource_group) > 0:
                        for resource_from_resource_group in resources_from_resource_group:
                            if isinstance(resource_from_resource_group, DockerResource):
                                resource_from_resource_group.set_workspace_settings(
                                    workspace_settings=self.workspace_settings
                                )
                                if resource_from_resource_group.group is None and self.name is not None:
                                    resource_from_resource_group.group = self.name
                                if resource_from_resource_group.should_delete(
                                    group_filter=group_filter,
                                    name_filter=name_filter,
                                    type_filter=type_filter,
                                ):
                                    resources_to_delete.append(resource_from_resource_group)
                elif isinstance(r, DockerResource):
                    r.set_workspace_settings(workspace_settings=self.workspace_settings)
                    if r.group is None and self.name is not None:
                        r.group = self.name
                    if r.should_delete(
                        group_filter=group_filter,
                        name_filter=name_filter,
                        type_filter=type_filter,
                    ):
                        r.set_workspace_settings(workspace_settings=self.workspace_settings)
                        resources_to_delete.append(r)

        # Build a list of DockerApps to delete
        apps_to_delete: List[DockerApp] = []
        if self.apps is not None:
            for app in self.apps:
                if isinstance(app, AppGroup):
                    apps_from_app_group = app.get_apps()
                    if len(apps_from_app_group) > 0:
                        for app_from_app_group in apps_from_app_group:
                            if isinstance(app_from_app_group, DockerApp):
                                if app_from_app_group.group is None and self.name is not None:
                                    app_from_app_group.group = self.name
                                if app_from_app_group.should_delete(group_filter=group_filter):
                                    apps_to_delete.append(app_from_app_group)
                elif isinstance(app, DockerApp):
                    if app.group is None and self.name is not None:
                        app.group = self.name
                    if app.should_delete(group_filter=group_filter):
                        apps_to_delete.append(app)

        # Get the list of DockerResources from the DockerApps
        if len(apps_to_delete) > 0:
            logger.debug(f"Found {len(apps_to_delete)} apps to delete")
            for app in apps_to_delete:
                app.set_workspace_settings(workspace_settings=self.workspace_settings)
                app_resources = app.get_resources(build_context=DockerBuildContext(network=self.network))
                if len(app_resources) > 0:
                    # Add the resources from the app to the list of resources to delete
                    for app_resource in app_resources:
                        if isinstance(app_resource, DockerResource) and app_resource.should_delete(
                            group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                        ):
                            resources_to_delete.append(app_resource)
                    # # If the app has dependencies, add the resources from the
                    # # dependencies to the list of resources to delete
                    # if app.depends_on is not None:
                    #     for dep in app.depends_on:
                    #         if isinstance(dep, DockerApp):
                    #             dep.set_workspace_settings(workspace_settings=self.workspace_settings)
                    #             dep_resources = dep.get_resources(
                    #                 build_context=DockerBuildContext(network=self.network)
                    #             )
                    #             if len(dep_resources) > 0:
                    #                 for dep_resource in dep_resources:
                    #                     if isinstance(dep_resource, DockerResource):
                    #                         resources_to_delete.append(dep_resource)

        # Sort the DockerResources in install order
        resources_to_delete.sort(key=lambda x: DockerResourceInstallOrder.get(x.__class__.__name__, 5000), reverse=True)

        # Deduplicate DockerResources
        deduped_resources_to_delete: List[DockerResource] = []
        for r in resources_to_delete:
            if r not in deduped_resources_to_delete:
                deduped_resources_to_delete.append(r)

        # Implement dependency sorting
        final_docker_resources: List[DockerResource] = []
        logger.debug("-*- Building DockerResources dependency graph")
        for docker_resource in deduped_resources_to_delete:
            # Logic to follow if resource has dependencies
            if docker_resource.depends_on is not None:
                # 1. Reverse the order of dependencies
                docker_resource.depends_on.reverse()

                # 2. Remove the dependencies if they are already added to the final_docker_resources
                for dep in docker_resource.depends_on:
                    if dep in final_docker_resources:
                        logger.debug(f"-*- Removing {dep.name}, dependency of {docker_resource.name}")
                        final_docker_resources.remove(dep)

                # 3. Add the resource to be deleted before its dependencies
                if docker_resource not in final_docker_resources:
                    logger.debug(f"-*- Adding {docker_resource.name}")
                    final_docker_resources.append(docker_resource)

                # 4. Add the dependencies back in reverse order
                for dep in docker_resource.depends_on:
                    if isinstance(dep, DockerResource):
                        if dep not in final_docker_resources:
                            logger.debug(f"-*- Adding {dep.name}, dependency of {docker_resource.name}")
                            final_docker_resources.append(dep)
            else:
                # Add the resource to be deleted if it has no dependencies
                if docker_resource not in final_docker_resources:
                    logger.debug(f"-*- Adding {docker_resource.name}")
                    final_docker_resources.append(docker_resource)

        # Track the total number of DockerResources to delete for validation
        num_resources_to_delete: int = len(final_docker_resources)
        num_resources_deleted: int = 0
        if num_resources_to_delete == 0:
            return 0, 0

        if dry_run:
            print_heading("--**- Docker resources to delete:")
            for resource in final_docker_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            print_info(f"\nNetwork: {self.network}")
            print_info(f"Total {num_resources_to_delete} resources")
            return 0, 0

        # Validate resources to be deleted
        if not auto_confirm:
            print_heading("\n--**-- Confirm resources to delete:")
            for resource in final_docker_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            print_info(f"\nNetwork: {self.network}")
            print_info(f"Total {num_resources_to_delete} resources")
            confirm = confirm_yes_no("\nConfirm delete")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping delete")
                print_info("-*-")
                return 0, 0

        for resource in final_docker_resources:
            print_info(f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}")
            if force is True:
                resource.force = True
            if isinstance(resource, DockerContainer):
                if resource.network is None and self.network is not None:
                    resource.network = self.network
            # logger.debug(resource)
            try:
                _resource_deleted = resource.delete(docker_client=self.docker_client)
                if _resource_deleted:
                    num_resources_deleted += 1
                else:
                    if self.workspace_settings is not None and not self.workspace_settings.continue_on_delete_failure:
                        return num_resources_deleted, num_resources_to_delete
            except Exception as e:
                logger.error(f"Failed to delete {resource.get_resource_type()}: {resource.get_resource_name()}")
                logger.error(e)
                logger.error("Please fix and try again...")

        print_heading(f"\n--**-- Resources deleted: {num_resources_deleted}/{num_resources_to_delete}")
        if num_resources_to_delete != num_resources_deleted:
            logger.error(
                f"Resources deleted: {num_resources_deleted} do not match resources required: {num_resources_to_delete}"
            )  # noqa: E501
        return num_resources_deleted, num_resources_to_delete

    def update_resources(
        self,
        group_filter: Optional[str] = None,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        dry_run: Optional[bool] = False,
        auto_confirm: Optional[bool] = False,
        force: Optional[bool] = None,
    ) -> Tuple[int, int]:
        from phi.cli.console import print_info, print_heading, confirm_yes_no
        from phi.docker.resource.types import DockerContainer, DockerResourceInstallOrder

        logger.debug("-*- Updating DockerResources")

        # Build a list of DockerResources to update
        resources_to_update: List[DockerResource] = []
        if self.resources is not None:
            for r in self.resources:
                if isinstance(r, ResourceGroup):
                    resources_from_resource_group = r.get_resources()
                    if len(resources_from_resource_group) > 0:
                        for resource_from_resource_group in resources_from_resource_group:
                            if isinstance(resource_from_resource_group, DockerResource):
                                resource_from_resource_group.set_workspace_settings(
                                    workspace_settings=self.workspace_settings
                                )
                                if resource_from_resource_group.group is None and self.name is not None:
                                    resource_from_resource_group.group = self.name
                                if resource_from_resource_group.should_update(
                                    group_filter=group_filter,
                                    name_filter=name_filter,
                                    type_filter=type_filter,
                                ):
                                    resources_to_update.append(resource_from_resource_group)
                elif isinstance(r, DockerResource):
                    r.set_workspace_settings(workspace_settings=self.workspace_settings)
                    if r.group is None and self.name is not None:
                        r.group = self.name
                    if r.should_update(
                        group_filter=group_filter,
                        name_filter=name_filter,
                        type_filter=type_filter,
                    ):
                        r.set_workspace_settings(workspace_settings=self.workspace_settings)
                        resources_to_update.append(r)

        # Build a list of DockerApps to update
        apps_to_update: List[DockerApp] = []
        if self.apps is not None:
            for app in self.apps:
                if isinstance(app, AppGroup):
                    apps_from_app_group = app.get_apps()
                    if len(apps_from_app_group) > 0:
                        for app_from_app_group in apps_from_app_group:
                            if isinstance(app_from_app_group, DockerApp):
                                if app_from_app_group.group is None and self.name is not None:
                                    app_from_app_group.group = self.name
                                if app_from_app_group.should_update(group_filter=group_filter):
                                    apps_to_update.append(app_from_app_group)
                elif isinstance(app, DockerApp):
                    if app.group is None and self.name is not None:
                        app.group = self.name
                    if app.should_update(group_filter=group_filter):
                        apps_to_update.append(app)

        # Get the list of DockerResources from the DockerApps
        if len(apps_to_update) > 0:
            logger.debug(f"Found {len(apps_to_update)} apps to update")
            for app in apps_to_update:
                app.set_workspace_settings(workspace_settings=self.workspace_settings)
                app_resources = app.get_resources(build_context=DockerBuildContext(network=self.network))
                if len(app_resources) > 0:
                    # # If the app has dependencies, add the resources from the
                    # # dependencies first to the list of resources to update
                    # if app.depends_on is not None:
                    #     for dep in app.depends_on:
                    #         if isinstance(dep, DockerApp):
                    #             dep.set_workspace_settings(workspace_settings=self.workspace_settings)
                    #             dep_resources = dep.get_resources(
                    #                 build_context=DockerBuildContext(network=self.network)
                    #             )
                    #             if len(dep_resources) > 0:
                    #                 for dep_resource in dep_resources:
                    #                     if isinstance(dep_resource, DockerResource):
                    #                         resources_to_update.append(dep_resource)
                    # Add the resources from the app to the list of resources to update
                    for app_resource in app_resources:
                        if isinstance(app_resource, DockerResource) and app_resource.should_update(
                            group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                        ):
                            resources_to_update.append(app_resource)

        # Sort the DockerResources in install order
        resources_to_update.sort(key=lambda x: DockerResourceInstallOrder.get(x.__class__.__name__, 5000), reverse=True)

        # Deduplicate DockerResources
        deduped_resources_to_update: List[DockerResource] = []
        for r in resources_to_update:
            if r not in deduped_resources_to_update:
                deduped_resources_to_update.append(r)

        # Implement dependency sorting
        final_docker_resources: List[DockerResource] = []
        logger.debug("-*- Building DockerResources dependency graph")
        for docker_resource in deduped_resources_to_update:
            # Logic to follow if resource has dependencies
            if docker_resource.depends_on is not None:
                # Add the dependencies before the resource itself
                for dep in docker_resource.depends_on:
                    if isinstance(dep, DockerResource):
                        if dep not in final_docker_resources:
                            logger.debug(f"-*- Adding {dep.name}, dependency of {docker_resource.name}")
                            final_docker_resources.append(dep)

                # Add the resource to be created after its dependencies
                if docker_resource not in final_docker_resources:
                    logger.debug(f"-*- Adding {docker_resource.name}")
                    final_docker_resources.append(docker_resource)
            else:
                # Add the resource to be created if it has no dependencies
                if docker_resource not in final_docker_resources:
                    logger.debug(f"-*- Adding {docker_resource.name}")
                    final_docker_resources.append(docker_resource)

        # Track the total number of DockerResources to update for validation
        num_resources_to_update: int = len(final_docker_resources)
        num_resources_updated: int = 0
        if num_resources_to_update == 0:
            return 0, 0

        if dry_run:
            print_heading("--**- Docker resources to update:")
            for resource in final_docker_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            print_info(f"\nNetwork: {self.network}")
            print_info(f"Total {num_resources_to_update} resources")
            return 0, 0

        # Validate resources to be updated
        if not auto_confirm:
            print_heading("\n--**-- Confirm resources to update:")
            for resource in final_docker_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            print_info(f"\nNetwork: {self.network}")
            print_info(f"Total {num_resources_to_update} resources")
            confirm = confirm_yes_no("\nConfirm patch")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping update")
                print_info("-*-")
                return 0, 0

        for resource in final_docker_resources:
            print_info(f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}")
            if force is True:
                resource.force = True
            if isinstance(resource, DockerContainer):
                if resource.network is None and self.network is not None:
                    resource.network = self.network
            # logger.debug(resource)
            try:
                _resource_updated = resource.update(docker_client=self.docker_client)
                if _resource_updated:
                    num_resources_updated += 1
                else:
                    if self.workspace_settings is not None and not self.workspace_settings.continue_on_patch_failure:
                        return num_resources_updated, num_resources_to_update
            except Exception as e:
                logger.error(f"Failed to update {resource.get_resource_type()}: {resource.get_resource_name()}")
                logger.error(e)
                logger.error("Please fix and try again...")

        print_heading(f"\n--**-- Resources updated: {num_resources_updated}/{num_resources_to_update}")
        if num_resources_to_update != num_resources_updated:
            logger.error(
                f"Resources updated: {num_resources_updated} do not match resources required: {num_resources_to_update}"
            )  # noqa: E501
        return num_resources_updated, num_resources_to_update

    def save_resources(
        self,
        group_filter: Optional[str] = None,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        workspace_settings: Optional[WorkspaceSettings] = None,
    ) -> Tuple[int, int]:
        raise NotImplementedError



from enum import Enum
from typing import Optional, Dict
from pathlib import Path

from phi.docker.app.base import DockerApp, ContainerContext  # noqa: F401
from phi.app.db_app import DbApp
from phi.utils.common import str_to_int
from phi.utils.log import logger


class AirflowLogsVolumeType(str, Enum):
    HostPath = "HostPath"
    EmptyDir = "EmptyDir"


class AirflowBase(DockerApp):
    # -*- App Name
    name: str = "airflow"

    # -*- Image Configuration
    image_name: str = "phidata/airflow"
    image_tag: str = "2.7.1"

    # -*- App Ports
    # Open a container port if open_port=True
    open_port: bool = False
    port_number: int = 8080

    # -*- Workspace Configuration
    # Path to the workspace directory inside the container
    workspace_dir_container_path: str = "/usr/local/workspace"
    # Mount the workspace directory from host machine to the container
    mount_workspace: bool = False

    # -*- Airflow Configuration
    # airflow_env sets the AIRFLOW_ENV env var and can be used by
    # DAGs to separate dev/stg/prd code
    airflow_env: Optional[str] = None
    # Set the AIRFLOW_HOME env variable
    # Defaults to: /usr/local/airflow
    airflow_home: Optional[str] = None
    # Set the AIRFLOW__CORE__DAGS_FOLDER env variable to the workspace_root/{airflow_dags_dir}
    # By default, airflow_dags_dir is set to the "dags" folder in the workspace
    airflow_dags_dir: str = "dags"
    # Creates an airflow admin with username: admin, pass: admin
    create_airflow_admin_user: bool = False
    # Airflow Executor
    executor: str = "SequentialExecutor"

    # -*- Airflow Database Configuration
    # Set as True to wait for db before starting airflow
    wait_for_db: bool = False
    # Set as True to delay start by 60 seconds so that the db can be initialized
    wait_for_db_init: bool = False
    # Connect to the database using a DbApp
    db_app: Optional[DbApp] = None
    # Provide database connection details manually
    # db_user can be provided here or as the
    # DB_USER env var in the secrets_file
    db_user: Optional[str] = None
    # db_password can be provided here or as the
    # DB_PASSWORD env var in the secrets_file
    db_password: Optional[str] = None
    # db_database can be provided here or as the
    # DB_DATABASE env var in the secrets_file
    db_database: Optional[str] = None
    # db_host can be provided here or as the
    # DB_HOST env var in the secrets_file
    db_host: Optional[str] = None
    # db_port can be provided here or as the
    # DB_PORT env var in the secrets_file
    db_port: Optional[int] = None
    # db_driver can be provided here or as the
    # DB_DRIVER env var in the secrets_file
    db_driver: str = "postgresql+psycopg2"
    db_result_backend_driver: str = "db+postgresql"
    # Airflow db connections in the format { conn_id: conn_url }
    # converted to env var: AIRFLOW_CONN__conn_id = conn_url
    db_connections: Optional[Dict] = None
    # Set as True to migrate (initialize/upgrade) the airflow_db
    db_migrate: bool = False

    # -*- Airflow Redis Configuration
    # Set as True to wait for redis before starting airflow
    wait_for_redis: bool = False
    # Connect to redis using a DbApp
    redis_app: Optional[DbApp] = None
    # Provide redis connection details manually
    # redis_password can be provided here or as the
    # REDIS_PASSWORD env var in the secrets_file
    redis_password: Optional[str] = None
    # redis_schema can be provided here or as the
    # REDIS_SCHEMA env var in the secrets_file
    redis_schema: Optional[str] = None
    # redis_host can be provided here or as the
    # REDIS_HOST env var in the secrets_file
    redis_host: Optional[str] = None
    # redis_port can be provided here or as the
    # REDIS_PORT env var in the secrets_file
    redis_port: Optional[int] = None
    # redis_driver can be provided here or as the
    # REDIS_DRIVER env var in the secrets_file
    redis_driver: str = "redis"

    # -*- Logs Volume
    # Mount the logs directory on the container
    mount_logs: bool = True
    logs_volume_name: Optional[str] = None
    logs_volume_type: AirflowLogsVolumeType = AirflowLogsVolumeType.EmptyDir
    # Container path to mount the volume
    # - If logs_volume_container_path is provided, use that
    # - If logs_volume_container_path is None and airflow_home is set
    #       use airflow_home/logs
    # - If logs_volume_container_path is None and airflow_home is None
    #       use "/usr/local/airflow/logs"
    logs_volume_container_path: Optional[str] = None
    # Host path to mount the postgres volume
    # If volume_type = PostgresVolumeType.HOST_PATH
    logs_volume_host_path: Optional[Path] = None

    #  -*- Other args
    load_examples: bool = False

    def get_db_user(self) -> Optional[str]:
        return self.db_user or self.get_secret_from_file("DB_USER")

    def get_db_password(self) -> Optional[str]:
        return self.db_password or self.get_secret_from_file("DB_PASSWORD")

    def get_db_database(self) -> Optional[str]:
        return self.db_database or self.get_secret_from_file("DB_DATABASE")

    def get_db_driver(self) -> Optional[str]:
        return self.db_driver or self.get_secret_from_file("DB_DRIVER")

    def get_db_host(self) -> Optional[str]:
        return self.db_host or self.get_secret_from_file("DB_HOST")

    def get_db_port(self) -> Optional[int]:
        return self.db_port or str_to_int(self.get_secret_from_file("DB_PORT"))

    def get_redis_password(self) -> Optional[str]:
        return self.redis_password or self.get_secret_from_file("REDIS_PASSWORD")

    def get_redis_schema(self) -> Optional[str]:
        return self.redis_schema or self.get_secret_from_file("REDIS_SCHEMA")

    def get_redis_host(self) -> Optional[str]:
        return self.redis_host or self.get_secret_from_file("REDIS_HOST")

    def get_redis_port(self) -> Optional[int]:
        return self.redis_port or str_to_int(self.get_secret_from_file("REDIS_PORT"))

    def get_redis_driver(self) -> Optional[str]:
        return self.redis_driver or self.get_secret_from_file("REDIS_DRIVER")

    def get_airflow_home(self) -> str:
        return self.airflow_home or "/usr/local/airflow"

    def get_container_env(self, container_context: ContainerContext) -> Dict[str, str]:
        from phi.constants import (
            PHI_RUNTIME_ENV_VAR,
            PYTHONPATH_ENV_VAR,
            REQUIREMENTS_FILE_PATH_ENV_VAR,
            SCRIPTS_DIR_ENV_VAR,
            STORAGE_DIR_ENV_VAR,
            WORKFLOWS_DIR_ENV_VAR,
            WORKSPACE_DIR_ENV_VAR,
            WORKSPACE_HASH_ENV_VAR,
            WORKSPACE_ID_ENV_VAR,
            WORKSPACE_ROOT_ENV_VAR,
            INIT_AIRFLOW_ENV_VAR,
            AIRFLOW_ENV_ENV_VAR,
            AIRFLOW_HOME_ENV_VAR,
            AIRFLOW_DAGS_FOLDER_ENV_VAR,
            AIRFLOW_EXECUTOR_ENV_VAR,
            AIRFLOW_DB_CONN_URL_ENV_VAR,
        )

        # Container Environment
        container_env: Dict[str, str] = self.container_env or {}
        container_env.update(
            {
                "INSTALL_REQUIREMENTS": str(self.install_requirements),
                "MOUNT_RESOURCES": str(self.mount_resources),
                "MOUNT_WORKSPACE": str(self.mount_workspace),
                "PRINT_ENV_ON_LOAD": str(self.print_env_on_load),
                "RESOURCES_DIR_CONTAINER_PATH": str(self.resources_dir_container_path),
                PHI_RUNTIME_ENV_VAR: "docker",
                REQUIREMENTS_FILE_PATH_ENV_VAR: container_context.requirements_file or "",
                SCRIPTS_DIR_ENV_VAR: container_context.scripts_dir or "",
                STORAGE_DIR_ENV_VAR: container_context.storage_dir or "",
                WORKFLOWS_DIR_ENV_VAR: container_context.workflows_dir or "",
                WORKSPACE_DIR_ENV_VAR: container_context.workspace_dir or "",
                WORKSPACE_ROOT_ENV_VAR: container_context.workspace_root or "",
                # Env variables used by Airflow
                "MOUNT_LOGS": str(self.mount_logs),
                # INIT_AIRFLOW env var is required for phidata to generate DAGs from workflows
                INIT_AIRFLOW_ENV_VAR: str(True),
                "DB_MIGRATE": str(self.db_migrate),
                "WAIT_FOR_DB": str(self.wait_for_db),
                "WAIT_FOR_DB_INIT": str(self.wait_for_db_init),
                "WAIT_FOR_REDIS": str(self.wait_for_redis),
                "CREATE_AIRFLOW_ADMIN_USER": str(self.create_airflow_admin_user),
                AIRFLOW_EXECUTOR_ENV_VAR: str(self.executor),
                "AIRFLOW__CORE__LOAD_EXAMPLES": str(self.load_examples),
            }
        )

        try:
            if container_context.workspace_schema is not None:
                if container_context.workspace_schema.id_workspace is not None:
                    container_env[WORKSPACE_ID_ENV_VAR] = str(container_context.workspace_schema.id_workspace) or ""
                if container_context.workspace_schema.ws_hash is not None:
                    container_env[WORKSPACE_HASH_ENV_VAR] = container_context.workspace_schema.ws_hash
        except Exception:
            pass

        if self.set_python_path:
            python_path = self.python_path
            if python_path is None:
                python_path = f"{container_context.workspace_root}:{self.get_airflow_home()}"
                if self.mount_resources and self.resources_dir_container_path is not None:
                    python_path = "{}:{}".format(python_path, self.resources_dir_container_path)
                if self.add_python_paths is not None:
                    python_path = "{}:{}".format(python_path, ":".join(self.add_python_paths))
            if python_path is not None:
                container_env[PYTHONPATH_ENV_VAR] = python_path

        # Set aws region and profile
        self.set_aws_env_vars(env_dict=container_env)

        # Set the AIRFLOW__CORE__DAGS_FOLDER
        container_env[AIRFLOW_DAGS_FOLDER_ENV_VAR] = f"{container_context.workspace_root}/{self.airflow_dags_dir}"

        # Set the AIRFLOW_ENV
        if self.airflow_env is not None:
            container_env[AIRFLOW_ENV_ENV_VAR] = self.airflow_env

        # Set the AIRFLOW_HOME
        if self.airflow_home is not None:
            container_env[AIRFLOW_HOME_ENV_VAR] = self.get_airflow_home()

        # Set the AIRFLOW__CONN_ variables
        if self.db_connections is not None:
            for conn_id, conn_url in self.db_connections.items():
                try:
                    af_conn_id = str("AIRFLOW_CONN_{}".format(conn_id)).upper()
                    container_env[af_conn_id] = conn_url
                except Exception as e:
                    logger.exception(e)
                    continue

        # Airflow db connection
        db_user = self.get_db_user()
        db_password = self.get_db_password()
        db_database = self.get_db_database()
        db_host = self.get_db_host()
        db_port = self.get_db_port()
        db_driver = self.get_db_driver()
        if self.db_app is not None and isinstance(self.db_app, DbApp):
            logger.debug(f"Reading db connection details from: {self.db_app.name}")
            if db_user is None:
                db_user = self.db_app.get_db_user()
            if db_password is None:
                db_password = self.db_app.get_db_password()
            if db_database is None:
                db_database = self.db_app.get_db_database()
            if db_host is None:
                db_host = self.db_app.get_db_host()
            if db_port is None:
                db_port = self.db_app.get_db_port()
            if db_driver is None:
                db_driver = self.db_app.get_db_driver()
        db_connection_url = f"{db_driver}://{db_user}:{db_password}@{db_host}:{db_port}/{db_database}"

        # Set the AIRFLOW__DATABASE__SQL_ALCHEMY_CONN
        if "None" not in db_connection_url:
            logger.debug(f"AIRFLOW__DATABASE__SQL_ALCHEMY_CONN: {db_connection_url}")
            container_env[AIRFLOW_DB_CONN_URL_ENV_VAR] = db_connection_url

        # Set the database connection details in the container env
        if db_host is not None:
            container_env["DATABASE_HOST"] = db_host
        if db_port is not None:
            container_env["DATABASE_PORT"] = str(db_port)

        # Airflow redis connection
        if self.executor == "CeleryExecutor":
            # Airflow celery result backend
            celery_result_backend_driver = self.db_result_backend_driver or db_driver
            celery_result_backend_url = (
                f"{celery_result_backend_driver}://{db_user}:{db_password}@{db_host}:{db_port}/{db_database}"
            )
            # Set the AIRFLOW__CELERY__RESULT_BACKEND
            if "None" not in celery_result_backend_url:
                container_env["AIRFLOW__CELERY__RESULT_BACKEND"] = celery_result_backend_url

            # Airflow celery broker url
            _redis_pass = self.get_redis_password()
            redis_password = f"{_redis_pass}@" if _redis_pass else ""
            redis_schema = self.get_redis_schema()
            redis_host = self.get_redis_host()
            redis_port = self.get_redis_port()
            redis_driver = self.get_redis_driver()
            if self.redis_app is not None and isinstance(self.redis_app, DbApp):
                logger.debug(f"Reading redis connection details from: {self.redis_app.name}")
                if redis_password is None:
                    redis_password = self.redis_app.get_db_password()
                if redis_schema is None:
                    redis_schema = self.redis_app.get_db_database() or "0"
                if redis_host is None:
                    redis_host = self.redis_app.get_db_host()
                if redis_port is None:
                    redis_port = self.redis_app.get_db_port()
                if redis_driver is None:
                    redis_driver = self.redis_app.get_db_driver()

            # Set the AIRFLOW__CELERY__RESULT_BACKEND
            celery_broker_url = f"{redis_driver}://{redis_password}{redis_host}:{redis_port}/{redis_schema}"
            if "None" not in celery_broker_url:
                logger.debug(f"AIRFLOW__CELERY__BROKER_URL: {celery_broker_url}")
                container_env["AIRFLOW__CELERY__BROKER_URL"] = celery_broker_url

            # Set the redis connection details in the container env
            if redis_host is not None:
                container_env["REDIS_HOST"] = redis_host
            if redis_port is not None:
                container_env["REDIS_PORT"] = str(redis_port)

        # Update the container env using env_file
        env_data_from_file = self.get_env_file_data()
        if env_data_from_file is not None:
            container_env.update({k: str(v) for k, v in env_data_from_file.items() if v is not None})

        # Update the container env using secrets_file
        secret_data_from_file = self.get_secret_file_data()
        if secret_data_from_file is not None:
            container_env.update({k: str(v) for k, v in secret_data_from_file.items() if v is not None})

        # Update the container env with user provided env_vars
        # this overwrites any existing variables with the same key
        if self.env_vars is not None and isinstance(self.env_vars, dict):
            container_env.update({k: str(v) for k, v in self.env_vars.items() if v is not None})

        # logger.debug("Container Environment: {}".format(container_env))
        return container_env

    def get_container_volumes(self, container_context: ContainerContext) -> Dict[str, dict]:
        from phi.utils.defaults import get_default_volume_name

        container_volumes: Dict[str, dict] = super().get_container_volumes(container_context=container_context)

        # Create Logs Volume
        if self.mount_logs:
            logs_volume_container_path_str = self.logs_volume_container_path
            if logs_volume_container_path_str is None:
                logs_volume_container_path_str = f"{self.get_airflow_home()}/logs"

            if self.logs_volume_type == AirflowLogsVolumeType.EmptyDir:
                logs_volume_name = self.logs_volume_name
                if logs_volume_name is None:
                    logs_volume_name = get_default_volume_name(f"{self.get_app_name()}-logs")
                logger.debug(f"Mounting: {logs_volume_name}")
                logger.debug(f"\tto: {logs_volume_container_path_str}")
                container_volumes[logs_volume_name] = {
                    "bind": logs_volume_container_path_str,
                    "mode": "rw",
                }
            elif self.logs_volume_type == AirflowLogsVolumeType.HostPath:
                if self.logs_volume_host_path is not None:
                    logs_volume_host_path_str = str(self.logs_volume_host_path)
                    logger.debug(f"Mounting: {logs_volume_host_path_str}")
                    logger.debug(f"\tto: {logs_volume_container_path_str}")
                    container_volumes[logs_volume_host_path_str] = {
                        "bind": logs_volume_container_path_str,
                        "mode": "rw",
                    }
                else:
                    logger.error("Airflow: logs_volume_host_path is None")
            else:
                logger.error(f"{self.logs_volume_type.value} not supported")

        return container_volumes



import json
from uuid import uuid4
from typing import List, Any, Optional, Dict, Iterator, Callable, cast, Union

from pydantic import BaseModel, ConfigDict, field_validator, model_validator, Field

from phi.document import Document
from phi.knowledge.base import KnowledgeBase
from phi.llm.base import LLM
from phi.llm.openai import OpenAIChat
from phi.llm.schemas import Message
from phi.llm.function.registry import FunctionRegistry
from phi.utils.format_str import remove_indent
from phi.utils.log import logger, set_log_level_to_debug
from phi.utils.timer import Timer


class Task(BaseModel):
    # -*- LLM to use for this task
    llm: LLM = OpenAIChat()

    # -*- Task settings
    # Task UUID
    id: Optional[str] = Field(None, validate_default=True)
    # Task name
    name: Optional[str] = None
    # Metadata associated with this task
    meta_data: Optional[Dict[str, Any]] = None

    # -*- Task Output
    output: Optional[Any] = None

    # -*- Task Knowledge Base
    knowledge_base: Optional[KnowledgeBase] = None
    # Add references from the knowledge base to the prompt sent to the LLM.
    add_references_to_prompt: bool = False

    # -*- Enable Function Calls
    # Makes the task autonomous.
    function_calls: bool = False
    # Add a list of default functions to the LLM
    default_functions: bool = True
    # Show function calls in LLM messages.
    show_function_calls: bool = False
    # A list of functions to add to the LLM.
    functions: Optional[List[Callable]] = None
    # A list of function registries to add to the LLM.
    function_registries: Optional[List[FunctionRegistry]] = None

    #
    # -*- Prompt Settings
    #
    # -*- System prompt: provide the system prompt as a string or using a function
    system_prompt: Optional[str] = None
    # Function to build the system prompt.
    # This function is provided the task as an argument
    #   and should return the system_prompt as a string.
    # Signature:
    # def system_prompt_function(task: Task) -> str:
    #    ...
    system_prompt_function: Optional[Callable[..., Optional[str]]] = None
    # If True, the task provides a default system prompt
    use_default_system_prompt: bool = True

    # -*- User prompt: provide the user prompt as a string or using a function
    # Note: this will ignore the input provided to the run function
    user_prompt: Optional[str] = None
    # Function to build the user prompt.
    # This function is provided the task and the input as arguments
    #   and should return the user_prompt as a string.
    # If add_references_to_prompt is True, then references are also provided as an argument.
    # Signature:
    # def custom_user_prompt_function(
    #     task: Task,
    #     message: Optional[str] = None,
    #     references: Optional[str] = None,
    # ) -> str:
    #     ...
    user_prompt_function: Optional[Callable[..., str]] = None
    # If True, the task provides a default user prompt
    use_default_user_prompt: bool = True
    # Function to build references for the default user_prompt
    # This function, if provided, is called when add_references_to_prompt is True
    # Signature:
    # def references(task: Task, query: str) -> Optional[str]:
    #     ...
    references_function: Optional[Callable[..., Optional[str]]] = None

    # If True, show debug logs
    debug_mode: bool = False

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("id", mode="before")
    def set_task_id(cls, v: Optional[str]) -> str:
        return v if v is not None else str(uuid4())

    @field_validator("debug_mode", mode="before")
    def set_log_level(cls, v: bool) -> bool:
        if v:
            set_log_level_to_debug()
            logger.debug("Debug logs enabled")
        return v

    @model_validator(mode="after")
    def add_functions_to_llm(self) -> "Task":
        if self.function_calls:
            if self.default_functions:
                default_func_list: List[Callable] = [self.search_knowledge_base]
                for func in default_func_list:
                    self.llm.add_function(func)

            # Add functions from self.functions
            if self.functions is not None:
                for func in self.functions:
                    self.llm.add_function(func)

            # Add functions from registries
            if self.function_registries is not None:
                for registry in self.function_registries:
                    self.llm.add_function_registry(registry)

            # Set function call to auto if it is not set
            if self.llm.function_call is None:
                self.llm.function_call = "auto"

            # Set show_function_calls if it is not set on the llm
            if self.llm.show_function_calls is None:
                self.llm.show_function_calls = self.show_function_calls
        return self

    def get_references_from_knowledge_base(self, query: str, num_documents: Optional[int] = None) -> Optional[str]:
        """Return a list of references from the knowledge base"""

        if self.references_function is not None:
            reference_kwargs = {"task": self, "query": query}
            return remove_indent(self.references_function(**reference_kwargs))

        if self.knowledge_base is None:
            return None

        relevant_docs: List[Document] = self.knowledge_base.search(query=query, num_documents=num_documents)
        return json.dumps([doc.to_dict() for doc in relevant_docs])

    def get_system_prompt(self) -> Optional[str]:
        """Return the system prompt for the task"""

        # If the system_prompt is set, return it
        if self.system_prompt is not None:
            return "\n".join([line.strip() for line in self.system_prompt.split("\n")])

        # If the system_prompt_function is set, return the system_prompt from the function
        if self.system_prompt_function is not None:
            system_prompt_kwargs = {"task": self}
            _system_prompt_from_function = remove_indent(self.system_prompt_function(**system_prompt_kwargs))
            if _system_prompt_from_function is not None:
                return _system_prompt_from_function
            else:
                raise Exception("system_prompt_function returned None")

        # If use_default_system_prompt is False, return None
        if not self.use_default_system_prompt:
            return None

        # Build a default system prompt
        _system_prompt = "You are a helpful assistant.\n"

        if self.knowledge_base is not None:
            _system_prompt += "You have access to a knowledge base that you can search for information.\n"

        if self.function_calls:
            _system_prompt += "You have access to functions that you can run to help you respond to the user.\n"

        # Return the system prompt after removing newlines and indenting
        _system_prompt = cast(str, remove_indent(_system_prompt))
        return _system_prompt

    def get_user_prompt(self, message: Optional[str] = None, references: Optional[str] = None) -> str:
        """Build the user prompt given a message and references"""

        # If the user_prompt is set, return it
        # Note: this ignores the message provided to the run function
        if self.user_prompt is not None:
            return "\n".join([line.strip() for line in self.user_prompt.split("\n")])

        # If the user_prompt_function is set, return the user_prompt from the function
        if self.user_prompt_function is not None:
            user_prompt_kwargs = {
                "task": self,
                "message": message,
                "references": references,
            }
            _user_prompt_from_function = remove_indent(self.user_prompt_function(**user_prompt_kwargs))
            if _user_prompt_from_function is not None:
                return _user_prompt_from_function
            else:
                raise Exception("user_prompt_function returned None")

        if message is None:
            raise Exception("Could not build user prompt. Please provide an input message.")

        # If use_default_user_prompt is False, return the message as is
        if not self.use_default_user_prompt:
            return message

        # If references and chat_history are None, return the message as is
        if references is None:
            return message

        # Build a default user prompt
        _user_prompt = "Respond to the following message in the best way possible.\n"

        # Add references to prompt
        if references:
            _user_prompt += f"""\n
                You can use the following information from the knowledge base if it helps respond to the message.
                START OF KNOWLEDGE BASE INFORMATION
                ```
                {references}
                ```
                END OF KNOWLEDGE BASE INFORMATION
                """

        # Remind the LLM of its task
        # if references:
        #     _user_prompt += "\nRemember, your task is to respond to the following message."

        _user_prompt += f"\nUSER: {message}"
        _user_prompt += "\nASSISTANT: "

        # Return the user prompt after removing newlines and indenting
        _user_prompt = cast(str, remove_indent(_user_prompt))
        return _user_prompt

    def _run(self, message: Optional[str] = None, stream: bool = True) -> Iterator[str]:
        logger.debug("*********** Task Run Start ***********")

        # -*- Build the system prompt
        system_prompt = self.get_system_prompt()

        # -*- Get references to add to the user_prompt
        user_prompt_references = None
        if self.add_references_to_prompt and message:
            reference_timer = Timer()
            reference_timer.start()
            user_prompt_references = self.get_references_from_knowledge_base(query=message)
            reference_timer.stop()
            logger.debug(f"Time to get references: {reference_timer.elapsed:.4f}s")

        # -*- Build the user prompt
        user_prompt = self.get_user_prompt(message=message, references=user_prompt_references)

        # -*- Build the messages to send to the LLM
        # Create system message
        system_prompt_message = Message(role="system", content=system_prompt)
        # Create user message
        user_prompt_message = Message(role="user", content=user_prompt)

        # Create message list
        messages: List[Message] = []
        if system_prompt_message.content and system_prompt_message.content != "":
            messages.append(system_prompt_message)
        messages += [user_prompt_message]

        # -*- Generate response (includes running function calls)
        llm_response = ""
        if stream:
            for response_chunk in self.llm.parsed_response_stream(messages=messages):
                llm_response += response_chunk
                yield response_chunk
        else:
            llm_response = self.llm.parsed_response(messages=messages)

        # -*- Yield final response if not streaming
        if not stream:
            yield llm_response
        logger.debug("*********** Task Run End ***********")

    def run(self, message: Optional[str] = None, stream: bool = True) -> Union[Iterator[str], str]:
        resp = self._run(message=message, stream=stream)
        if stream:
            return resp
        else:
            return next(resp)

    ###########################################################################
    # LLM functions
    ###########################################################################

    def search_knowledge_base(self, query: str) -> Optional[str]:
        """Search the knowledge base for information about a users query.

        :param query: The query to search for.
        :return: A string containing the response from the knowledge base.
        """
        return self.get_references_from_knowledge_base(query=query)

    ###########################################################################
    # Print Response
    ###########################################################################

    def print_response(self, message: str, stream: bool = True) -> None:
        from phi.cli.console import console
        from rich.live import Live
        from rich.table import Table
        from rich.box import ROUNDED
        from rich.markdown import Markdown

        if stream:
            response = ""
            with Live() as live_log:
                response_timer = Timer()
                response_timer.start()
                for resp in self.run(message, stream=True):
                    response += resp

                    table = Table(box=ROUNDED, border_style="blue")
                    table.add_column("Message")
                    table.add_column(message)
                    md_response = Markdown(response)
                    table.add_row(f"Response\n({response_timer.elapsed:.1f}s)", md_response)
                    live_log.update(table)
                response_timer.stop()
        else:
            response_timer = Timer()
            response_timer.start()
            response = self.run(message, stream=False)  # type: ignore
            response_timer.stop()
            md_response = Markdown(response)
            table = Table(box=ROUNDED, border_style="blue")
            table.add_column("Message")
            table.add_column(message)
            table.add_row(f"Response\n({response_timer.elapsed:.1f}s)", md_response)
            console.print(table)



from typing import Iterator, List

from phi.document import Document
from phi.document.reader.arxiv import ArxivReader
from phi.knowledge.base import KnowledgeBase


class ArxivKnowledgeBase(KnowledgeBase):
    queries: List[str] = []
    reader: ArxivReader = ArxivReader()

    @property
    def document_lists(self) -> Iterator[List[Document]]:
        """Iterate over urls and yield lists of documents.
        Each object yielded by the iterator is a list of documents.

        Returns:
            Iterator[List[Document]]: Iterator yielding list of documents
        """

        for _query in self.queries:
            yield self.reader.read(query=_query)



import json
from typing import Optional, List, Iterator, Dict, Any

from phi.aws.api_client import AwsApiClient
from phi.llm.base import LLM
from phi.llm.schemas import Message
from phi.utils.log import logger
from phi.utils.timer import Timer

try:
    from boto3 import session  # noqa: F401
except ImportError:
    logger.error("`boto3` not installed")
    raise


class AwsBedrock(LLM):
    name: str = "AwsBedrock"
    model: str = "anthropic.claude-v2"

    aws_region: Optional[str] = None
    aws_profile: Optional[str] = None
    aws_client: Optional[AwsApiClient] = None

    _bedrock_client: Optional[Any] = None
    _bedrock_runtime_client: Optional[Any] = None

    def get_aws_region(self) -> Optional[str]:
        # Priority 1: Use aws_region from model
        if self.aws_region is not None:
            return self.aws_region

        # Priority 2: Get aws_region from env
        from os import getenv
        from phi.constants import AWS_REGION_ENV_VAR

        aws_region_env = getenv(AWS_REGION_ENV_VAR)
        if aws_region_env is not None:
            self.aws_region = aws_region_env
        return self.aws_region

    def get_aws_profile(self) -> Optional[str]:
        # Priority 1: Use aws_region from resource
        if self.aws_profile is not None:
            return self.aws_profile

        # Priority 2: Get aws_profile from env
        from os import getenv
        from phi.constants import AWS_PROFILE_ENV_VAR

        aws_profile_env = getenv(AWS_PROFILE_ENV_VAR)
        if aws_profile_env is not None:
            self.aws_profile = aws_profile_env
        return self.aws_profile

    def get_aws_client(self) -> AwsApiClient:
        if self.aws_client is not None:
            return self.aws_client

        self.aws_client = AwsApiClient(aws_region=self.get_aws_region(), aws_profile=self.get_aws_profile())
        return self.aws_client

    @property
    def bedrock_client(self):
        if self._bedrock_client is not None:
            return self._bedrock_client

        boto3_session: session = self.get_aws_client().boto3_session
        self._bedrock_client = boto3_session.client(service_name="bedrock")
        return self._bedrock_client

    @property
    def bedrock_runtime_client(self):
        if self._bedrock_runtime_client is not None:
            return self._bedrock_runtime_client

        boto3_session: session = self.get_aws_client().boto3_session
        self._bedrock_runtime_client = boto3_session.client(service_name="bedrock-runtime")
        return self._bedrock_runtime_client

    @property
    def api_kwargs(self) -> Dict[str, Any]:
        return {}

    def get_model_summaries(self) -> List[Dict[str, Any]]:
        list_response: dict = self.bedrock_client.list_foundation_models()
        if list_response is None or "modelSummaries" not in list_response:
            return []

        return list_response["modelSummaries"]

    def get_model_ids(self) -> List[str]:
        model_summaries: List[Dict[str, Any]] = self.get_model_summaries()
        if len(model_summaries) == 0:
            return []

        return [model_summary["modelId"] for model_summary in model_summaries]

    def get_model_details(self) -> Dict[str, Any]:
        model_details: dict = self.bedrock_client.get_foundation_model(modelIdentifier=self.model)

        if model_details is None or "modelDetails" not in model_details:
            return {}

        return model_details["modelDetails"]

    def invoke_model(self, prompt: str) -> Dict[str, Any]:
        body = {"prompt": prompt}
        body.update(self.api_kwargs)
        body_json = json.dumps(body)

        response = self.bedrock_runtime_client.invoke_model(
            body=body_json,
            modelId="anthropic.claude-v2",
            accept="application/json",
            contentType="application/json",
        )

        response_body = response.get("body")
        if response_body is None:
            return {}

        return json.loads(response_body.read())

    def invoke_model_stream(self, prompt: str) -> Iterator[Dict[str, Any]]:
        body = {"prompt": prompt}
        body.update(self.api_kwargs)
        body_json = json.dumps(body)

        response = self.bedrock_runtime_client.invoke_model_with_response_stream(
            body=body_json,
            modelId="anthropic.claude-v2",
        )

        for event in response.get("body"):
            chunk = event.get("chunk")
            if chunk:
                yield json.loads(chunk.get("bytes").decode())

    def build_prompt(self, messages: List[Message]) -> str:
        """Build prompt from messages"""

        prompt = ""
        for m in messages:
            if m.role == "user":
                if m.content is not None:
                    prompt += m.content

        # -*- Log prompt for debugging
        logger.debug(f"PROMPT: {prompt}")
        return prompt

    def parsed_response(self, messages: List[Message]) -> str:
        logger.debug("---------- Aws Response Start ----------")

        response_timer = Timer()
        response_timer.start()
        response: Dict[str, Any] = self.invoke_model(prompt=self.build_prompt(messages))
        response_timer.stop()
        logger.debug(f"Time to generate response: {response_timer.elapsed:.4f}s")

        # -*- Parse response
        response_completion = response.get("completion")

        # -*- Create assistant message
        assistant_message = Message(
            role="assistant",
            content=response_completion,
        )

        # -*- Update usage metrics
        # Add response time to metrics
        assistant_message.metrics["time"] = response_timer.elapsed
        if "response_times" not in self.metrics:
            self.metrics["response_times"] = []
        self.metrics["response_times"].append(response_timer.elapsed)

        # Add token usage to metrics
        prompt_tokens = 0
        if prompt_tokens is not None:
            assistant_message.metrics["prompt_tokens"] = prompt_tokens
            if "prompt_tokens" not in self.metrics:
                self.metrics["prompt_tokens"] = prompt_tokens
            else:
                self.metrics["prompt_tokens"] += prompt_tokens
        completion_tokens = 0
        if completion_tokens is not None:
            assistant_message.metrics["completion_tokens"] = completion_tokens
            if "completion_tokens" not in self.metrics:
                self.metrics["completion_tokens"] = completion_tokens
            else:
                self.metrics["completion_tokens"] += completion_tokens
        total_tokens = prompt_tokens + completion_tokens
        if total_tokens is not None:
            assistant_message.metrics["total_tokens"] = total_tokens
            if "total_tokens" not in self.metrics:
                self.metrics["total_tokens"] = total_tokens
            else:
                self.metrics["total_tokens"] += total_tokens

        # -*- Add assistant message to messages
        messages.append(assistant_message)
        assistant_message.log()

        logger.debug("---------- Aws Response End ----------")
        # -*- Return content
        return assistant_message.content if assistant_message.content is not None else ""

    def response_message(self, messages: List[Message]) -> Dict:
        logger.debug("---------- Aws Response Start ----------")

        response_timer = Timer()
        response_timer.start()
        response: Dict[str, Any] = self.invoke_model(prompt=self.build_prompt(messages))
        response_timer.stop()
        logger.debug(f"Time to generate response: {response_timer.elapsed:.4f}s")

        # -*- Parse response
        response_completion = response.get("completion")

        # -*- Create assistant message
        assistant_message = Message(
            role="assistant",
            content=response_completion,
        )

        # -*- Update usage metrics
        # Add response time to metrics
        assistant_message.metrics["time"] = response_timer.elapsed
        if "response_times" not in self.metrics:
            self.metrics["response_times"] = []
        self.metrics["response_times"].append(response_timer.elapsed)

        # Add token usage to metrics
        prompt_tokens = 0
        if prompt_tokens is not None:
            assistant_message.metrics["prompt_tokens"] = prompt_tokens
            if "prompt_tokens" not in self.metrics:
                self.metrics["prompt_tokens"] = prompt_tokens
            else:
                self.metrics["prompt_tokens"] += prompt_tokens
        completion_tokens = 0
        if completion_tokens is not None:
            assistant_message.metrics["completion_tokens"] = completion_tokens
            if "completion_tokens" not in self.metrics:
                self.metrics["completion_tokens"] = completion_tokens
            else:
                self.metrics["completion_tokens"] += completion_tokens
        total_tokens = prompt_tokens + completion_tokens
        if total_tokens is not None:
            assistant_message.metrics["total_tokens"] = total_tokens
            if "total_tokens" not in self.metrics:
                self.metrics["total_tokens"] = total_tokens
            else:
                self.metrics["total_tokens"] += total_tokens

        # -*- Add assistant message to messages
        messages.append(assistant_message)
        assistant_message.log()

        logger.debug("---------- Aws Response End ----------")
        # -*- Return content
        return response

    def parsed_response_stream(self, messages: List[Message]) -> Iterator[str]:
        logger.debug("---------- Aws Response Start ----------")

        assistant_message_content = ""
        completion_tokens = 0
        response_timer = Timer()
        response_timer.start()
        for delta in self.invoke_model_stream(prompt=self.build_prompt(messages)):
            completion_tokens += 1
            # -*- Parse response
            delta_completion = delta.get("completion")
            # -*- Yield completion
            if delta_completion is not None:
                assistant_message_content += delta_completion
                yield delta_completion

        response_timer.stop()
        logger.debug(f"Time to generate response: {response_timer.elapsed:.4f}s")

        # -*- Create assistant message
        assistant_message = Message(role="assistant")
        # -*- Add content to assistant message
        if assistant_message_content != "":
            assistant_message.content = assistant_message_content

        # -*- Update usage metrics
        # Add response time to metrics
        assistant_message.metrics["time"] = response_timer.elapsed
        if "response_times" not in self.metrics:
            self.metrics["response_times"] = []
        self.metrics["response_times"].append(response_timer.elapsed)

        # Add token usage to metrics
        prompt_tokens = 0
        assistant_message.metrics["prompt_tokens"] = prompt_tokens
        if "prompt_tokens" not in self.metrics:
            self.metrics["prompt_tokens"] = prompt_tokens
        else:
            self.metrics["prompt_tokens"] += prompt_tokens
        logger.debug(f"Estimated completion tokens: {completion_tokens}")
        assistant_message.metrics["completion_tokens"] = completion_tokens
        if "completion_tokens" not in self.metrics:
            self.metrics["completion_tokens"] = completion_tokens
        else:
            self.metrics["completion_tokens"] += completion_tokens
        total_tokens = prompt_tokens + completion_tokens
        assistant_message.metrics["total_tokens"] = total_tokens
        if "total_tokens" not in self.metrics:
            self.metrics["total_tokens"] = total_tokens
        else:
            self.metrics["total_tokens"] += total_tokens

        # -*- Add assistant message to messages
        messages.append(assistant_message)
        assistant_message.log()
        logger.debug("---------- Aws Response End ----------")

    def response_delta(self, messages: List[Message]) -> Iterator[Dict]:
        logger.debug("---------- OpenAI Response Start ----------")

        assistant_message_content = ""
        completion_tokens = 0
        response_timer = Timer()
        response_timer.start()
        for delta in self.invoke_model_stream(prompt=self.build_prompt(messages)):
            completion_tokens += 1
            # -*- Parse response
            delta_completion = delta.get("completion")
            # -*- Yield completion
            if delta_completion is not None:
                assistant_message_content += delta_completion
                yield delta

        response_timer.stop()
        logger.debug(f"Time to generate response: {response_timer.elapsed:.4f}s")

        # -*- Create assistant message
        assistant_message = Message(role="assistant")
        # -*- Add content to assistant message
        if assistant_message_content != "":
            assistant_message.content = assistant_message_content

        # -*- Update usage metrics
        # Add response time to metrics
        assistant_message.metrics["time"] = response_timer.elapsed
        if "response_times" not in self.metrics:
            self.metrics["response_times"] = []
        self.metrics["response_times"].append(response_timer.elapsed)

        # Add token usage to metrics
        prompt_tokens = 0
        assistant_message.metrics["prompt_tokens"] = prompt_tokens
        if "prompt_tokens" not in self.metrics:
            self.metrics["prompt_tokens"] = prompt_tokens
        else:
            self.metrics["prompt_tokens"] += prompt_tokens
        logger.debug(f"Estimated completion tokens: {completion_tokens}")
        assistant_message.metrics["completion_tokens"] = completion_tokens
        if "completion_tokens" not in self.metrics:
            self.metrics["completion_tokens"] = completion_tokens
        else:
            self.metrics["completion_tokens"] += completion_tokens
        total_tokens = prompt_tokens + completion_tokens
        assistant_message.metrics["total_tokens"] = total_tokens
        if "total_tokens" not in self.metrics:
            self.metrics["total_tokens"] = total_tokens
        else:
            self.metrics["total_tokens"] += total_tokens

        # -*- Add assistant message to messages
        messages.append(assistant_message)
        assistant_message.log()
        logger.debug("---------- Aws Response End ----------")



from typing import Iterator, List

from phi.document import Document
from phi.knowledge.base import KnowledgeBase

try:
    import wikipedia  # noqa: F401
except ImportError:
    raise ImportError("The `wikipedia` package is not installed. Please install it via `pip install wikipedia`.")


class WikipediaKnowledgeBase(KnowledgeBase):
    topics: List[str] = []

    @property
    def document_lists(self) -> Iterator[List[Document]]:
        """Iterate over urls and yield lists of documents.
        Each object yielded by the iterator is a list of documents.

        Returns:
            Iterator[List[Document]]: Iterator yielding list of documents
        """

        for topic in self.topics:
            yield [
                Document(
                    name=topic,
                    meta_data={"topic": topic},
                    content=wikipedia.summary(topic),
                )
            ]



from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import Field, BaseModel, ConfigDict, field_serializer

from phi.resource.base import ResourceBase
from phi.k8s.api_client import K8sApiClient
from phi.k8s.constants import DEFAULT_K8S_NAMESPACE
from phi.k8s.enums.api_version import ApiVersion
from phi.k8s.enums.kind import Kind
from phi.k8s.resource.meta.v1.object_meta import ObjectMeta
from phi.cli.console import print_info
from phi.utils.log import logger


class K8sObject(BaseModel):
    def get_k8s_object(self) -> Any:
        """Creates a K8sObject for this resource.
        Eg:
            * For a Deployment resource, it will return the V1Deployment object.
        """
        logger.error("@get_k8s_object method not defined")

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)


class K8sResource(ResourceBase, K8sObject):
    """Base class for K8s Resources"""

    # Common fields for all K8s Resources
    # Which version of the Kubernetes API you're using to create this object
    # Note: we use an alias "apiVersion" so that the K8s manifest generated by this resource
    #       has the correct key
    api_version: ApiVersion = Field(..., alias="apiVersion")
    # What kind of object you want to create
    kind: Kind
    # Data that helps uniquely identify the object, including a name string, UID, and optional namespace
    metadata: ObjectMeta

    # Fields used in api calls
    # async_req bool: execute request asynchronously
    async_req: bool = False
    # pretty: If 'true', then the output is pretty printed.
    pretty: bool = True

    # List of fields to include from the K8sResource base class when generating the
    # K8s manifest. Subclasses should add fields to the fields_for_k8s_manifest list to include them in the manifest.
    fields_for_k8s_manifest_base: List[str] = [
        "api_version",
        "apiVersion",
        "kind",
        "metadata",
    ]
    # List of fields to include from Subclasses when generating the K8s manifest.
    # This should be defined by the Subclass
    fields_for_k8s_manifest: List[str] = []

    k8s_client: Optional[K8sApiClient] = None

    @field_serializer("api_version")
    def get_api_version_value(self, v) -> str:
        return v.value

    @field_serializer("kind")
    def get_kind_value(self, v) -> str:
        return v.value

    def get_resource_name(self) -> str:
        return self.name or self.metadata.name or self.__class__.__name__

    def get_namespace(self) -> str:
        if self.metadata and self.metadata.namespace:
            return self.metadata.namespace
        return DEFAULT_K8S_NAMESPACE

    def get_label_selector(self) -> str:
        labels = self.metadata.labels
        if labels:
            label_str = ",".join([f"{k}={v}" for k, v in labels.items()])
            return label_str
        return ""

    @staticmethod
    def get_from_cluster(k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs) -> Any:
        """Gets all resources of this type from the k8s cluster"""
        logger.error("@get_from_cluster method not defined")
        return None

    def get_k8s_client(self) -> K8sApiClient:
        if self.k8s_client is not None:
            return self.k8s_client
        self.k8s_client = K8sApiClient()
        return self.k8s_client

    def _read(self, k8s_client: K8sApiClient) -> Any:
        logger.error(f"@_read method not defined for {self.get_resource_name()}")
        return True

    def read(self, k8s_client: K8sApiClient) -> Any:
        """Reads the resource from the k8s cluster
        Eg:
            * For a Deployment resource, it will return the V1Deployment object
            currently running on the cluster.
        """
        # Step 1: Use cached value if available
        if self.use_cache and self.active_resource is not None:
            return self.active_resource

        # Step 2: Skip resource creation if skip_read = True
        if self.skip_read:
            print_info(f"Skipping read: {self.get_resource_name()}")
            return True

        # Step 3: Read resource
        client: K8sApiClient = k8s_client or self.get_k8s_client()
        return self._read(client)

    def is_active(self, k8s_client: K8sApiClient) -> bool:
        """Returns True if the resource is active on the k8s cluster"""
        self.active_resource = self._read(k8s_client=k8s_client)
        return True if self.active_resource is not None else False

    def _create(self, k8s_client: K8sApiClient) -> bool:
        logger.error(f"@_create method not defined for {self.get_resource_name()}")
        return True

    def create(self, k8s_client: K8sApiClient) -> bool:
        """Creates the resource on the k8s Cluster"""

        # Step 1: Skip resource creation if skip_create = True
        if self.skip_create:
            print_info(f"Skipping create: {self.get_resource_name()}")
            return True

        # Step 2: Check if resource is active and use_cache = True
        client: K8sApiClient = k8s_client or self.get_k8s_client()
        if self.use_cache and self.is_active(client):
            self.resource_created = True
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} already exists")
            return True
        # Step 3: Create the resource
        else:
            self.resource_created = self._create(client)
            if self.resource_created:
                print_info(f"{self.get_resource_type()}: {self.get_resource_name()} created")

        # Step 4: Run post create steps
        if self.resource_created:
            if self.save_output:
                self.save_output_file()
            logger.debug(f"Running post-create for {self.get_resource_type()}: {self.get_resource_name()}")
            return self.post_create(client)
        logger.error(f"Failed to create {self.get_resource_type()}: {self.get_resource_name()}")
        return self.resource_created

    def post_create(self, k8s_client: K8sApiClient) -> bool:
        return True

    def _update(self, k8s_client: K8sApiClient) -> Any:
        logger.error(f"@_update method not defined for {self.get_resource_name()}")
        return True

    def update(self, k8s_client: K8sApiClient) -> bool:
        """Updates the resource on the k8s Cluster"""

        # Step 1: Skip resource update if skip_update = True
        if self.skip_update:
            print_info(f"Skipping update: {self.get_resource_name()}")
            return True

        # Step 2: Update the resource
        client: K8sApiClient = k8s_client or self.get_k8s_client()
        if self.is_active(client):
            self.resource_updated = self._update(client)
        else:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} does not exist")
            return True

        # Step 3: Run post update steps
        if self.resource_updated:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} updated")
            if self.save_output:
                self.save_output_file()
            logger.debug(f"Running post-update for {self.get_resource_type()}: {self.get_resource_name()}")
            return self.post_update(client)
        logger.error(f"Failed to update {self.get_resource_type()}: {self.get_resource_name()}")
        return self.resource_updated

    def post_update(self, k8s_client: K8sApiClient) -> bool:
        return True

    def _delete(self, k8s_client: K8sApiClient) -> Any:
        logger.error(f"@_delete method not defined for {self.get_resource_name()}")
        return True

    def delete(self, k8s_client: K8sApiClient) -> bool:
        """Deletes the resource from the k8s cluster"""

        # Step 1: Skip resource deletion if skip_delete = True
        if self.skip_delete:
            print_info(f"Skipping delete: {self.get_resource_name()}")
            return True

        # Step 2: Delete the resource
        client: K8sApiClient = k8s_client or self.get_k8s_client()
        if self.is_active(client):
            self.resource_deleted = self._delete(client)
        else:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} does not exist")
            return True

        # Step 3: Run post delete steps
        if self.resource_deleted:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} deleted")
            if self.save_output:
                self.delete_output_file()
            logger.debug(f"Running post-delete for {self.get_resource_type()}: {self.get_resource_name()}.")
            return self.post_delete(client)
        logger.error(f"Failed to delete {self.get_resource_type()}: {self.get_resource_name()}")
        return self.resource_deleted

    def post_delete(self, k8s_client: K8sApiClient) -> bool:
        return True

    ######################################################
    ## Function to get the k8s manifest
    ######################################################

    def get_k8s_manifest_dict(self) -> Optional[Dict[str, Any]]:
        """Returns the K8s Manifest for this Object as a dict"""

        from itertools import chain

        k8s_manifest: Dict[str, Any] = {}
        all_attributes: Dict[str, Any] = self.model_dump(exclude_defaults=True, by_alias=True, exclude_none=True)
        # logger.debug("All Attributes: {}".format(all_attributes))
        for attr_name in chain(self.fields_for_k8s_manifest_base, self.fields_for_k8s_manifest):
            if attr_name in all_attributes:
                k8s_manifest[attr_name] = all_attributes[attr_name]
        # logger.debug(f"k8s_manifest:\n{k8s_manifest}")
        return k8s_manifest

    def get_k8s_manifest_yaml(self, **kwargs) -> Optional[str]:
        """Returns the K8s Manifest for this Object as a yaml"""

        import yaml

        k8s_manifest_dict = self.get_k8s_manifest_dict()

        if k8s_manifest_dict is not None:
            return yaml.safe_dump(k8s_manifest_dict, **kwargs)
        return None

    def get_k8s_manifest_json(self, **kwargs) -> Optional[str]:
        """Returns the K8s Manifest for this Object as a json"""

        import json

        k8s_manifest_dict = self.get_k8s_manifest_dict()

        if k8s_manifest_dict is not None:
            return json.dumps(k8s_manifest_dict, **kwargs)
        return None

    def save_manifests(self, **kwargs) -> Optional[Path]:
        """Saves the K8s Manifests for this Object to the input file

        Returns:
            Path: The path to the input file
        """
        input_file_path: Optional[Path] = self.get_input_file_path()
        if input_file_path is None:
            return None

        input_file_path_parent: Optional[Path] = input_file_path.parent
        # Create parent directory if needed
        if input_file_path_parent is not None and not input_file_path_parent.exists():
            input_file_path_parent.mkdir(parents=True, exist_ok=True)

        manifest_yaml = self.get_k8s_manifest_yaml(**kwargs)
        if manifest_yaml is not None:
            logger.debug(f"Writing {str(input_file_path)}")
            input_file_path.write_text(manifest_yaml)
            return input_file_path
        return None



from phi.k8s.app.postgres.postgres import PostgresDb


class PgVectorDb(PostgresDb):
    # -*- App Name
    name: str = "pgvector"

    # -*- Image Configuration
    image_name: str = "phidata/pgvector"
    image_tag: str = "15"



from typing import Optional

from pydantic import BaseModel


class UserSchema(BaseModel):
    """Schema for user data returned by the API."""

    id_user: int
    email: Optional[str] = None
    username: Optional[str] = None
    is_active: Optional[bool] = True
    is_bot: Optional[bool] = False
    name: Optional[str] = None
    email_verified: Optional[bool] = False


class EmailPasswordAuthSchema(BaseModel):
    email: str
    password: str
    auth_source: str = "cli"



from typing import Union, Optional
from typing_extensions import Literal

from kubernetes.client.models.v1_deployment_strategy import V1DeploymentStrategy
from kubernetes.client.models.v1_rolling_update_deployment import (
    V1RollingUpdateDeployment,
)
from pydantic import Field

from phi.k8s.resource.base import K8sObject


class RollingUpdateDeployment(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#rollingupdatedeployment-v1-apps
    """

    resource_type: str = "RollingUpdateDeployment"

    max_surge: Optional[Union[int, str]] = Field(None, alias="maxSurge")
    max_unavailable: Optional[Union[int, str]] = Field(None, alias="maxUnavailable")

    def get_k8s_object(self) -> V1RollingUpdateDeployment:
        # Return a V1RollingUpdateDeployment object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_rolling_update_deployment.py
        _v1_rolling_update_deployment = V1RollingUpdateDeployment(
            max_surge=self.max_surge,
            max_unavailable=self.max_unavailable,
        )
        return _v1_rolling_update_deployment


class DeploymentStrategy(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#deploymentstrategy-v1-apps
    """

    resource_type: str = "DeploymentStrategy"

    rolling_update: RollingUpdateDeployment = Field(RollingUpdateDeployment(), alias="rollingUpdate")
    type: Literal["Recreate", "RollingUpdate"] = "RollingUpdate"

    def get_k8s_object(self) -> V1DeploymentStrategy:
        # Return a V1DeploymentStrategy object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_deployment_strategy.py
        _v1_deployment_strategy = V1DeploymentStrategy(
            rolling_update=self.rolling_update.get_k8s_object(),
            type=self.type,
        )
        return _v1_deployment_strategy



from typing import Any, Dict, Optional

from pydantic import BaseModel


class WorkspaceCreate(BaseModel):
    ws_name: str
    git_url: Optional[str] = None
    is_primary_for_user: Optional[bool] = False
    visibility: Optional[str] = None
    ws_data: Optional[Dict[str, Any]] = None


class WorkspaceUpdate(BaseModel):
    id_workspace: int
    ws_name: Optional[str] = None
    git_url: Optional[str] = None
    visibility: Optional[str] = None
    ws_data: Optional[Dict[str, Any]] = None
    is_active: Optional[bool] = None


class UpdatePrimaryWorkspace(BaseModel):
    id_workspace: int
    ws_name: Optional[str] = None


class WorkspaceDelete(BaseModel):
    id_workspace: int
    ws_name: Optional[str] = None


class WorkspaceEvent(BaseModel):
    id_workspace: int
    event_type: str
    event_status: str
    event_data: Optional[Dict[str, Any]] = None


class WorkspaceSchema(BaseModel):
    """Workspace data returned by the API."""

    id_workspace: Optional[int] = None
    ws_name: Optional[str] = None
    is_active: Optional[bool] = None
    git_url: Optional[str] = None
    ws_hash: Optional[str] = None
    ws_data: Optional[Dict[str, Any]] = None



from typing import List, Optional

from kubernetes.client import CoreV1Api
from kubernetes.client.models.v1_service_account import V1ServiceAccount
from kubernetes.client.models.v1_service_account_list import V1ServiceAccountList
from pydantic import Field

from phi.k8s.api_client import K8sApiClient
from phi.k8s.resource.core.v1.local_object_reference import (
    LocalObjectReference,
)
from phi.k8s.resource.core.v1.object_reference import ObjectReference
from phi.k8s.resource.base import K8sResource
from phi.utils.log import logger


class ServiceAccount(K8sResource):
    """A service account provides an identity for processes that run in a Pod.
    When you create a pod, if you do not specify a service account, it is automatically assigned the default
    service account in the same namespace.

    References:
    - Docs: https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#serviceaccount-v1-core
    - Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_service_account.py
    """

    resource_type: str = "ServiceAccount"

    # AutomountServiceAccountToken indicates whether pods running as this service account
    # should have an API token automatically mounted. Can be overridden at the pod level.
    automount_service_account_token: Optional[bool] = Field(None, alias="automountServiceAccountToken")
    # ImagePullSecrets is a list of references to secrets in the same namespace to use for pulling any images in pods
    # that reference this ServiceAccount. ImagePullSecrets are distinct from Secrets because Secrets can be mounted
    # in the pod, but ImagePullSecrets are only accessed by the kubelet.
    # More info: https://kubernetes.io/docs/concepts/containers/images/#specifying-imagepullsecrets-on-a-pod
    image_pull_secrets: Optional[List[LocalObjectReference]] = Field(None, alias="imagePullSecrets")
    # Secrets is the list of secrets allowed to be used by pods running using this ServiceAccount.
    # More info: https://kubernetes.io/docs/concepts/configuration/secret
    secrets: Optional[List[ObjectReference]] = None

    # List of attributes to include in the K8s manifest
    fields_for_k8s_manifest: List[str] = [
        "automount_service_account_token",
        "image_pull_secrets",
        "secrets",
    ]

    def get_k8s_object(self) -> V1ServiceAccount:
        """Creates a body for this ServiceAccount"""

        # Return a V1ServiceAccount object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_service_account.py
        _image_pull_secrets = None
        if self.image_pull_secrets:
            _image_pull_secrets = []
            for ips in self.image_pull_secrets:
                _image_pull_secrets.append(ips.get_k8s_object())

        _secrets = None
        if self.secrets:
            _secrets = []
            for s in self.secrets:
                _secrets.append(s.get_k8s_object())

        _v1_service_account = V1ServiceAccount(
            api_version=self.api_version.value,
            kind=self.kind.value,
            metadata=self.metadata.get_k8s_object(),
            automount_service_account_token=self.automount_service_account_token,
            image_pull_secrets=_image_pull_secrets,
            secrets=_secrets,
        )
        return _v1_service_account

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs
    ) -> Optional[List[V1ServiceAccount]]:
        """Reads ServiceAccounts from K8s cluster.

        Args:
            k8s_client: K8sApiClient for the cluster
            namespace: Namespace to use.
        """
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        sa_list: Optional[V1ServiceAccountList] = None
        if namespace:
            # logger.debug(f"Getting SAs for ns: {namespace}")
            sa_list = core_v1_api.list_namespaced_service_account(namespace=namespace)
        else:
            # logger.debug("Getting SAs for all namespaces")
            sa_list = core_v1_api.list_service_account_for_all_namespaces()

        sas: Optional[List[V1ServiceAccount]] = None
        if sa_list:
            sas = sa_list.items
        # logger.debug(f"sas: {sas}")
        # logger.debug(f"sas type: {type(sas)}")

        return sas

    def _create(self, k8s_client: K8sApiClient) -> bool:
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        k8s_object: V1ServiceAccount = self.get_k8s_object()
        namespace = self.get_namespace()

        logger.debug("Creating: {}".format(self.get_resource_name()))
        v1_service_account: V1ServiceAccount = core_v1_api.create_namespaced_service_account(
            body=k8s_object,
            namespace=namespace,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Created: {}".format(v1_service_account))
        if v1_service_account.metadata.creation_timestamp is not None:
            logger.debug("ServiceAccount Created")
            self.active_resource = v1_service_account
            return True
        logger.error("ServiceAccount could not be created")
        return False

    def _read(self, k8s_client: K8sApiClient) -> Optional[V1ServiceAccount]:
        """Returns the "Active" ServiceAccount from the cluster"""

        namespace = self.get_namespace()
        active_resource: Optional[V1ServiceAccount] = None
        active_resources: Optional[List[V1ServiceAccount]] = self.get_from_cluster(
            k8s_client=k8s_client,
            namespace=namespace,
        )
        # logger.debug(f"Active Resources: {active_resources}")
        if active_resources is None:
            return None

        active_resources_dict = {_sa.metadata.name: _sa for _sa in active_resources}

        sa_name = self.get_resource_name()
        if sa_name in active_resources_dict:
            active_resource = active_resources_dict[sa_name]
            self.active_resource = active_resource
            logger.debug(f"Found active {sa_name}")
        return active_resource

    def _update(self, k8s_client: K8sApiClient) -> bool:
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        sa_name = self.get_resource_name()
        k8s_object: V1ServiceAccount = self.get_k8s_object()
        namespace = self.get_namespace()
        logger.debug("Updating: {}".format(sa_name))

        v1_service_account: V1ServiceAccount = core_v1_api.patch_namespaced_service_account(
            name=sa_name,
            namespace=namespace,
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Updated:\n{}".format(pformat(v1_service_account.to_dict(), indent=2)))
        if v1_service_account.metadata.creation_timestamp is not None:
            logger.debug("ServiceAccount Updated")
            self.active_resource = v1_service_account
            return True
        logger.error("ServiceAccount could not be updated")
        return False

    def _delete(self, k8s_client: K8sApiClient) -> bool:
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        sa_name = self.get_resource_name()
        namespace = self.get_namespace()

        logger.debug("Deleting: {}".format(sa_name))
        self.active_resource = None
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_status.py
        delete_status: V1ServiceAccount = core_v1_api.delete_namespaced_service_account(
            name=sa_name,
            namespace=namespace,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("delete_status: {}".format(delete_status))
        # logger.debug("delete_status type: {}".format(type(delete_status)))
        # logger.debug("delete_status: {}".format(delete_status.status))
        # TODO: validate the delete status, this check is currently not accurate
        # it just checks if a V1ServiceAccount object was returned
        if delete_status is not None:
            logger.debug("ServiceAccount Deleted")
            return True
        logger.error("ServiceAccount could not be deleted")
        return False



from typing import Optional, Union, List

from phi.docker.app.airflow.base import AirflowBase


class AirflowScheduler(AirflowBase):
    # -*- App Name
    name: str = "airflow-scheduler"

    # Command for the container
    command: Optional[Union[str, List[str]]] = "scheduler"
